
html.wait, html.wait * {
    cursor: wait !important;
}

div.dropdown input {
    width: 100%;
    color: #555555;
}

div.dropdown input,
div.dropdown input:hover,
div.dropdown input:active,
div.dropdown input:focus {
    border-width: 0;
    outline: 0;
}

.nav {
  margin-bottom: 8px;
}

.navbar {
  margin-bottom: 10px;
}

input.dropdown-toggle {
    border-width: 0;
    padding: 0;
}

.typeahead-li {
    width: 100%;
}

.control-label.required:after {
    content: " *";
    color: red;
}

.form-horizontal .control-group {
    margin-bottom: 10px;
}

.form-horizontal .control-group:first-child {
    margin-top: 10px;
}

.form-horizontal .control-group.label-size1 .control-label {
    width: 80px;
}

.form-horizontal .control-group.label-size1 .controls {
    margin-left: 100px;
}

.form-horizontal .control-group.label-size2 .control-label {
    width: 120px;
}

.form-horizontal .control-group.label-size2 .controls {
    margin-left: 140px;
}

.form-horizontal .control-group.label-size4 .control-label {
    width: 240px;
}

.form-horizontal .control-group.label-size4 .controls {
    margin-left: 260px;
}

.form-horizontal .control-group.label-size5 .control-label {
    width: 320px;
}

.form-horizontal .control-group.label-size5 .controls {
    margin-left: 340px;
}

.modal {
    width: 600px;
}

.modal.edit-form,
.form-frame.edit-form {
    width: 600px;
}

.form-body .well,
.modal-body .well {
/*
    background-color: inherit;
*/
    min-height: 20px;
    width: auto;
    margin-bottom: 6px;
    padding: 6px;
}

.modal.fade {
    -webkit-transition: opacity 0.1s linear;
       -moz-transition: opacity 0.1s linear;
         -o-transition: opacity 0.1s linear;
            transition: opacity 0.1s linear;
}

.modal-backdrop,
.modal-backdrop.fade.in{
    opacity: 0.2;
    filter: alpha(opacity=20);
    background: #fff;
}

.small-pager {
    display: inline-block;
    font-size: 11.844px;
    font-weight: bold;
    line-height: 14px;
    white-space: nowrap;
    vertical-align: baseline;
    opacity: 0.7;
    filter: alpha(opacity=70);
}

#pager form {
    padding: 4px;
}

.pager-input {
    text-align: center;
}

.typeahead,
.popover {
    z-index: 3000;
}

.badge {
    padding-left: 5px;
    padding-right: 5px;
}

.help-badge {
    margin-left: 5px;
    cursor: pointer;
}

.help-badge-divider {
    margin-left: 14px;
 }

.expanded-btn {
    min-width: 120px;
}

input.input-integer {
  width: 26%;
  max-width: 110px;
}

input.input-float {
  width: 37%;
  max-width: 146px;
}

input.input-currency {
  width: 37%;
  max-width: 146px;
}

.datetime-input-container {
    width: 59%;
}

.date-input-container {
    width: 49%;
}

.lookupfield-input-container {
    width: 80%;
}

.lookupvalues-input-container {
    width: 66%;
}

input.input-lookupitem,
input.input-lookupvalues,
input.input-date,
input.input-datetime,
input.input-file {
    width: calc(100% - 2 * 40px);
    width: -moz-calc(100% - 2 * 40px);
    width: -webkit-calc(100% - 2 * 40px);
    width: -o-calc(100% - 2 * 40px);
}

input.input-file2 {
    width: calc(100% - 3 * 40px + 5px);
    width: -moz-calc(100% - 3 * 40px + 5px);
    width: -webkit-calc(100% - 3 * 40px + 5px);
    width: -o-calc(100% - 3 * 40px + 5px);
}

input.input-file3 {
    width: calc(100% - 4 * 40px + 7px);
    width: -moz-calc(100% - 4 * 40px + 7px);
    width: -webkit-calc(100% - 4 * 40px + 7px);
    width: -o-calc(100% - 4 * 40px + 7px);
}

input.input-lookupitem,
input.input-lookupvalues {
    min-width: 80px;
}

input.input-date {
    min-width: 69px;
    max-width: 110px;
}

input.input-datetime {
    min-width: 103px;
    max-width: 150px;
}

input.input-text {
  width: 80%;
}

.input-longtext {
  width: 80%;
}

@media (max-width: 480px) {
    .form-horizontal .control-group.label-size1 .controls,
    .form-horizontal .control-group.label-size2 .controls,
    .form-horizontal .control-group.label-size4 .controls,
    .form-horizontal .control-group.label-size5 .controls {
        margin-left: 0;
    }
}

@media (max-width: 767px) {
    .datetime-input-container {
        width: 54%;
    }

    .date-input-container {
        width: 46%;
    }

    .lookupfield-input-container {
        width: 80%;
    }

    .lookupvalues-input-container {
        width: 66%;
    }

    input.input-lookupitem,
    input.input-lookupvalues,
    input.input-date,
    input.input-datetime {
        width: calc(100% - 2 * 40px);
        width: -moz-calc(100% - 2 * 40px);
        width: -webkit-calc(100% - 2 * 40px);
        width: -o-calc(100% - 2 * 40px);
    }

    input.input-file {
        width: calc(100% - 3 * 40px + 5px);
        width: -moz-calc(100% - 3 * 40px + 5px);
        width: -webkit-calc(100% - 3 * 40px + 5px);
        width: -o-calc(100% - 3 * 40px + 5px);
    }
}

.template {
    display: none;
}

.inline {
    display: inline;
}

.close {
    font-size: 12px;
}


.dbtable tr th:first-child,
.dbtable tr td:first-child {
    border-left-style: none;
}


/*
.dbtable th.fake-column,
.dbtable td.fake-column {
    width: auto;
    border-left-style: none;
    padding: 0;
    margin: 0;
}
*/

.dbtable td,
.dbtable th {
    cursor: default;
    -webkit-user-select: none;
    -moz-user-select: none;
    -khtml-user-select: none;
    -ms-user-select: none;
}

.tooltip {
  z-index: 3000;
}

.tooltip-inner {
  max-width: 400px;
  padding: 8px;
  text-align: center;
  text-decoration: none;
  -webkit-border-radius: 4px;
     -moz-border-radius: 4px;
          border-radius: 4px;
}

.top-margin {
  margin-top: 6px;
}

.error-modal-border {
    border: 1px solid red;
}

.selection-box {
    position:    absolute;
    background: transparent;
    border:     1px dotted #999;
}

.dbtable thead th {
    vertical-align: middle;
}

.dbtable td.multi-select {
    width: 4%;
}

.dbtable th .multi-select .sel-count {
    margin-top: 4px;
    margin-bottom: 0;
    font-size: 10px;
    font-style: italic;
    color: #3a87ad;
}

.dbtable th .multi-select .sel-count.selected-shown {
    color: #b94a48;
}

.dbtree {
    padding: 2px;
    outline: none;
}

.dbtree li div {
    white-space: nowrap;
}

.dbtree li div {
    cursor: default;
    margin: 1px;
    -webkit-user-select: none;
    -moz-user-select: none;
    -ms-user-select: none;
    -o-user-select: none;
    -khtml-user-select: none;
    user-select: none;
}


.dbtree li.selected > div span.tree-text {
    background-color: #e5e5e5;
}

.dbtree li.selected-focused > div span.tree-text {
    background-color: #d9edf7;
}

.dbtree li div span.tree-text {
    display: inline-block;
    cursor: default;
    padding: 3px 8px;
}

.dbtree ul:first-child {
    margin-left: 0px;
}

.mselect-btn {
    padding: 3px 4px;
}

.bullet {
    margin-left: 12px;
    margin-top: 0px;
    padding-top: 0px;
}

.empty-bullet {
    margin-left: 26px;
}

div.th-container {
    overflow: hidden;
}

div.th-table {
    display: table;
    width: 100%;
    height: 100%;
}

div.th-text {
    display: table-cell;
    vertical-align: middle;
}

.accordion-toggle.history-toggle {
    outline: 0;
}

.accordion-heading.history-heading a:focus,
.accordion-heading.history-heading a:hover {
    text-decoration: none;
    outline: 0;
}

.accordion-group.history-group {
    border-color: #cceeff;
}

.form-body {
  position: relative;
  overflow-y: auto;
  padding: 14px 15px;
}

.no-border .form-body {
    padding: 0;
}

.form-header {
  font-size: 0;
  padding: 14px 15px 0px;
  margin-bottom: 0;
  *zoom: 1;
}

.no-border .form-header {
    padding: 8px 15px 14px;
}

.form-header:before,
.form-header:after {
  display: table;
  line-height: 0;
  content: "";
}

.form-header:after {
  clear: both;
}

.form-header .btn + .btn,
.form-header .btn + .btn-group {
  margin-bottom: 0;
  margin-left: 5px;
}

.form-footer {
  padding: 14px 15px 15px;
  margin-bottom: 0;
  text-align: right;
  -webkit-border-radius: 0 0 6px 6px;
     -moz-border-radius: 0 0 6px 6px;
          border-radius: 0 0 6px 6px;
  *zoom: 1;
}

.form-footer:before,
.form-footer:after {
  display: table;
  line-height: 0;
  content: "";
}

.form-footer:after {
  clear: both;
}

.form-footer .btn + .btn,
.form-footer .btn-group + .btn {
  margin-bottom: 0;
  margin-left: 5px;
}

.form-footer .btn-group .btn + .btn {
  margin-left: -1px;
}

.form-footer .btn-block + .btn-block {
  margin-left: 0;
}

.form-header .form-inline,
.form-footer .form-inline {
    margin-bottom: 0;
    padding: 0 9px;
    *zoom: 1;
}

#app-title {
    font-size: 20px;
    margin: 4px 0;
}

#app-user {
    margin-top: 14px;
    text-align: right;
}

.logging-info {
    margin-top: 10px;
    text-align: right;
}

#log-out {
    margin-left: 4px;
}

.close-tab-btn:hover,
#refresh-btn:hover,
#history-btn:hover  {
    background-color: #d9edf7;
}

.form-frame {
  border: 1px solid #ddd;
  border: 1px solid rgba(0, 0, 0, 0.1);
  *border: 1px solid #ddd;
  -webkit-border-radius: 4px;
     -moz-border-radius: 4px;
          border-radius: 4px;
  outline: none;
  -webkit-box-shadow: 0 3px 7px rgba(0, 0, 0, 0.1);
     -moz-box-shadow: 0 3px 7px rgba(0, 0, 0, 0.1);
          box-shadow: 0 3px 7px rgba(0, 0, 0, 0.1);
  -webkit-background-clip: padding-box;
     -moz-background-clip: padding-box;
          background-clip: padding-box;
}

.form-frame.no-border {
    border-style: none;
}

.modal-header {
    display: flex;
    flex-direction: row;
    align-items: center;
    justify-content: space-between;
    padding: 4px 15px;
}

.no-border .modal-header {
    border-style: none;
}

.modal .modal-header {
    padding: 14px 15px;
}

.header-close-btn,
.header-print-btn {
    align-self: flex-start
}

.header-filters {
    margin-right: 20px;
    flex-grow: 5;
    display: flex;
    align-items: center;
    font-weight: bold;
}

.header-search {
    flex-grow: 0;
}

#search-form {
    margin-bottom: 0;
    padding: 0;
}

#search-form .dropdown-menu > li > a {
    font-size: 12px;
    padding: 2px 20px;
}

#search-form #search-input {
    margin-bottom: 0;
}

.modal-title {
    margin: 10px 0;
}

.filters-text {
    font-weight: normal;
    padding: 0;
    margin-left: 4px;
    font-size: 13px;
}

.header-selected {
    font-weight: bold;
}

.flex-container {
}

sup {
    vertical-align: super;
    font-size: smaller;
}

.header-btn {
    padding: 3px 7px;
    margin-left: 6px;
}

.alert-absolute {
    position: absolute;
    z-index: 3000;
    display: block;
    min-width: 200px;
}

.paginate-scroll {
    opacity: 0.8;
    filter: alpha(opacity=80);
}

.pulsate {
    text-shadow: 0 -1px rgba(0, 0, 0, .1);
    -webkit-animation: pulsate 1.2s linear infinite;
    animation: pulsate 1.2s linear infinite;
}

@-webkit-keyframes pulsate {
 50% { color: #fff; text-shadow: 0 -1px rgba(0, 0, 0, .3), 0 0 5px #ffd, 0 0 8px #fff; }
}
@keyframes pulsate {
 50% { color: #fff; text-shadow: 0 -1px rgba(0, 0, 0, .3), 0 0 5px #ffd, 0 0 8px #fff; }
}

.view-detail {
    margin-top: 4px;
}

.logging-info {
    margin-top: 20px;
    text-align: right;
}
