
#task-tabs.nav {
    margin-bottom: 10px;
}

#task-tabs li a {
    color: #555555;
}

#task-tabs li.active a {
    color: #0088cc;
}

#content {
    border: solid 1px #eee;
    padding: 6px;
    -webkit-border-radius: 4px;
        -moz-border-radius: 4px;
            border-radius: 4px;
}

.modal {
    width: 560px;
}

.dbtree li.selected-focused > div span.tree-text {
    color: #333333;
    background-color: #d9edf7;
}

/*.table tbody tr.selected > td:not(.field-selected) {*/
/*    color: #333333;*/
/*    background-color: #e6f7ff;*/
/*}*/

/*.table tbody tr.selected-focused > td:not(.field-selected-focused) {*/
/*    color: #333333;*/
/*    background-color: #cceeff;*/
/*}*/

/*.table tbody td.field-selected {*/
/*    color: #333333;*/
/*    background-color: #b3e6ff*/
/*}*/

/*.table tbody td.field-selected-focused {*/
/*    color: #333333;*/
/*    background-color: #99ddff;*/
/*}*/

.title-panel {
  padding: 14px 15px 15px;
  margin-bottom: 0;
  text-align: right;
  background-color: #f5f5f5;

  border-bottom: 1px solid #ddd;
  -webkit-border-radius: 6px 6px 0 0;
     -moz-border-radius: 6px 6px 0 0;
          border-radius: 6px 6px 0 0;
  *zoom: 1;
  -webkit-box-shadow: inset 0 1px 0 #ffffff;
     -moz-box-shadow: inset 0 1px 0 #ffffff;
          box-shadow: inset 0 1px 0 #ffffff;
}

.title-panel:before,
.title-panel:after {
  display: table;
  line-height: 0;
  content: "";
}

.title-panel:after {
  clear: both;
}

.title-panel .btn + .btn {
  margin-bottom: 0;
  margin-left: 5px;
}

.title-panel .btn-group .btn + .btn {
  margin-left: -1px;
}

.title-panel .btn-block + .btn-block {
  margin-left: 0;
}

.vert-btn {
    margin-top: 1px;
    margin-bottom: 8px;
    width: 160px;
}

.table-fixed {
    width:100%;
    border-collapse:collapse;
    table-layout:fixed;
}

.td-panel {
    vertical-align: top;
}

#btns-panel {
    text-align: center;
}

.side-panel {
    padding: 10px 2px;
    border: solid 1px #eee;
    display: none;
}

.btns-divider {
    margin-top: 16px;
}

.fields-grid .inner-table td,
.langs-table .inner-table td {
    border: 0;
}

#left-box, #right-box {
    border: solid 1px #eee;
}

.sys_langs-view #right-box {
    border-style: none;
}

#left-box .nav-tabs > li > a {
  padding-right: 8px;
  padding-left: 8px;
}

/*
#left-box .dbtree {
    font-family: Monaco, Menlo, "Ubuntu Mono", Consolas, source-code-pro, monospace;
}
*/

#left-box .dbtree li div span.tree-text {
    padding: 1px 4px;
}

#right-box {
    padding-left: 10px;
}

.left-right-btn {
    margin-left: 11px;
}

.up-down-btn {
    margin-bottom: 6px;
}

#vert-btns-box {
    margin-top: 14px;
    text-align: center;
}

#left-btn {
    margin-top: 145px;
}

.arrow_btn {
    padding: 4px 6px;
}

#right-btn {
    margin-top: 3px;
}

#center-box {
    padding: 0px;
}

#editor-box {
    border: solid 1px #eee;
    margin-left: 10px;
    position: relative;
}

#editor {
    font-size: 14px;
    position: absolute;
    top: 0;
    right: 0;
    bottom: 0;
    left: 0;
}

#info-grids {
    border: solid 1px #eee;
}

#editor-tabs > .nav-tabs {
    margin-bottom: 4px;
}

.close-editor-btn:hover  {
    background-color: #d9edf7;
}

.code_error {
    font-size: 14px;
    margin-left: 10px;
    color: red;
}

.editor-title {
    color: #0088cc;
    margin-left: 2px;
    margin-top: 0;
    margin-bottom: 0;
}

.header-search {
    margin-bottom: 4px;
}

p.search {
    margin: 0px;
    font-family: 'Courier New', Courier, monospace;
    white-space: nowrap;
}

@media print {
    p.search {
        margin: 0px;
        padding: 0px;
        line-height: 50%;
    }
}

#options-box .control-label {
    margin: 0;
    margin-bottom: 5px;
}

.edit-title {
    margin-bottom: 12px;
}

.lang-warning {
    text-align: center;
    font-size: 14px;
    color: #0088cc;
}

p.lang-warning a {
    color: #e41e06;
    font-weight: bold;
}

.topic-badge {
    margin-left: 2px;
}

.topic-badge a {
    color: #ffffff;
    font-weight: bold;
    cursor: help;
}







